-- SQLBench-DS query 91 derived from TPC-DS query 91 under the terms of the TPC Fair Use Policy.
-- TPC-DS queries are Copyright 2021 Transaction Processing Performance Council.
-- This query was generated at scale factor 1.
select
  cc_call_center_id Call_Center,
  cc_name Call_Center_Name,
  cc_manager Manager,
  sum(cr_net_loss) Returns_Loss
from
  call_center,
  catalog_returns,
  date_dim,
  customer,
  customer_address,
  customer_demographics,
  household_demographics
where
  cr_call_center_sk = cc_call_center_sk
  and cr_returned_date_sk = d_date_sk
  and cr_returning_customer_sk = c_customer_sk
  and cd_demo_sk = c_current_cdemo_sk
  and hd_demo_sk = c_current_hdemo_sk
  and ca_address_sk = c_current_addr_sk
  and d_year = 2002
  and d_moy = 11
  and (
    (
      cd_marital_status = 'M'
      and cd_education_status = 'Unknown'
    )
    or (
      cd_marital_status = 'W'
      and cd_education_status = 'Advanced Degree'
    )
  )
  and hd_buy_potential like 'Unknown%'
  and ca_gmt_offset = -6
group by
  cc_call_center_id,
  cc_name,
  cc_manager,
  cd_marital_status,
  cd_education_status
order by
  sum(cr_net_loss) desc;
