-- SQLBench-DS query 7 derived from TPC-DS query 7 under the terms of the TPC Fair Use Policy.
-- TPC-DS queries are Copyright 2021 Transaction Processing Performance Council.
-- This query was generated at scale factor 1.
select
  i_item_id,
  avg(ss_quantity) agg1,
  avg(ss_list_price) agg2,
  avg(ss_coupon_amt) agg3,
  avg(ss_sales_price) agg4
from
  store_sales,
  customer_demographics,
  date_dim,
  item,
  promotion
where
  ss_sold_date_sk = d_date_sk
  and ss_item_sk = i_item_sk
  and ss_cdemo_sk = cd_demo_sk
  and ss_promo_sk = p_promo_sk
  and cd_gender = 'M'
  and cd_marital_status = 'M'
  and cd_education_status = '4 yr Degree'
  and (
    p_channel_email = 'N'
    or p_channel_event = 'N'
  )
  and d_year = 2001
group by
  i_item_id
order by
  i_item_id
LIMIT
  100;
