-- SQLBench-DS query 41 derived from TPC-DS query 41 under the terms of the TPC Fair Use Policy.
-- TPC-DS queries are Copyright 2021 Transaction Processing Performance Council.
-- This query was generated at scale factor 1.
select distinct
  (i_product_name)
from
  item i1
where
  i_manufact_id between 668 and 668  + 40
  and (
    select
      count(*) as item_cnt
    from
      item
    where
      (
        i_manufact = i1.i_manufact
        and (
          (
            i_category = 'Women'
            and (
              i_color = 'cream'
              or i_color = 'ghost'
            )
            and (
              i_units = 'Ton'
              or i_units = 'Gross'
            )
            and (
              i_size = 'economy'
              or i_size = 'small'
            )
          )
          or (
            i_category = 'Women'
            and (
              i_color = 'midnight'
              or i_color = 'burlywood'
            )
            and (
              i_units = 'Tsp'
              or i_units = 'Bundle'
            )
            and (
              i_size = 'medium'
              or i_size = 'extra large'
            )
          )
          or (
            i_category = 'Men'
            and (
              i_color = 'lavender'
              or i_color = 'azure'
            )
            and (
              i_units = 'Each'
              or i_units = 'Lb'
            )
            and (
              i_size = 'large'
              or i_size = 'N/A'
            )
          )
          or (
            i_category = 'Men'
            and (
              i_color = 'chocolate'
              or i_color = 'steel'
            )
            and (
              i_units = 'N/A'
              or i_units = 'Dozen'
            )
            and (
              i_size = 'economy'
              or i_size = 'small'
            )
          )
        )
      )
      or (
        i_manufact = i1.i_manufact
        and (
          (
            i_category = 'Women'
            and (
              i_color = 'floral'
              or i_color = 'royal'
            )
            and (
              i_units = 'Unknown'
              or i_units = 'Tbl'
            )
            and (
              i_size = 'economy'
              or i_size = 'small'
            )
          )
          or (
            i_category = 'Women'
            and (
              i_color = 'navy'
              or i_color = 'forest'
            )
            and (
              i_units = 'Bunch'
              or i_units = 'Dram'
            )
            and (
              i_size = 'medium'
              or i_size = 'extra large'
            )
          )
          or (
            i_category = 'Men'
            and (
              i_color = 'cyan'
              or i_color = 'indian'
            )
            and (
              i_units = 'Carton'
              or i_units = 'Cup'
            )
            and (
              i_size = 'large'
              or i_size = 'N/A'
            )
          )
          or (
            i_category = 'Men'
            and (
              i_color = 'coral'
              or i_color = 'pale'
            )
            and (
              i_units = 'Pallet'
              or i_units = 'Gram'
            )
            and (
              i_size = 'economy'
              or i_size = 'small'
            )
          )
        )
      )
  ) > 0
order by
  i_product_name
LIMIT
  100;
