-- SQLBench-DS query 25 derived from TPC-DS query 25 under the terms of the TPC Fair Use Policy.
-- TPC-DS queries are Copyright 2021 Transaction Processing Performance Council.
-- This query was generated at scale factor 1.
select
  i_item_id,
  i_item_desc,
  s_store_id,
  s_store_name,
  min(ss_net_profit) as store_sales_profit,
  min(sr_net_loss) as store_returns_loss,
  min(cs_net_profit) as catalog_sales_profit
from
  store_sales,
  store_returns,
  catalog_sales,
  date_dim d1,
  date_dim d2,
  date_dim d3,
  store,
  item
where
  d1.d_moy = 4
  and d1.d_year = 2002
  and d1.d_date_sk = ss_sold_date_sk
  and i_item_sk = ss_item_sk
  and s_store_sk = ss_store_sk
  and ss_customer_sk = sr_customer_sk
  and ss_item_sk = sr_item_sk
  and ss_ticket_number = sr_ticket_number
  and sr_returned_date_sk = d2.d_date_sk
  and d2.d_moy between 4 and 10
  and d2.d_year = 2002
  and sr_customer_sk = cs_bill_customer_sk
  and sr_item_sk = cs_item_sk
  and cs_sold_date_sk = d3.d_date_sk
  and d3.d_moy between 4 and 10
  and d3.d_year = 2002
group by
  i_item_id,
  i_item_desc,
  s_store_id,
  s_store_name
order by
  i_item_id,
  i_item_desc,
  s_store_id,
  s_store_name
LIMIT
  100;
