-- SQLBench-DS query 38 derived from TPC-DS query 38 under the terms of the TPC Fair Use Policy.
-- TPC-DS queries are Copyright 2021 Transaction Processing Performance Council.
-- This query was generated at scale factor 1.
select
  count(*)
from
  (
    select distinct
      c_last_name,
      c_first_name,
      d_date
    from
      store_sales,
      date_dim,
      customer
    where
      store_sales.ss_sold_date_sk = date_dim.d_date_sk
      and store_sales.ss_customer_sk = customer.c_customer_sk
      and d_month_seq between 1190 and 1190  + 11
    intersect
    select distinct
      c_last_name,
      c_first_name,
      d_date
    from
      catalog_sales,
      date_dim,
      customer
    where
      catalog_sales.cs_sold_date_sk = date_dim.d_date_sk
      and catalog_sales.cs_bill_customer_sk = customer.c_customer_sk
      and d_month_seq between 1190 and 1190  + 11
    intersect
    select distinct
      c_last_name,
      c_first_name,
      d_date
    from
      web_sales,
      date_dim,
      customer
    where
      web_sales.ws_sold_date_sk = date_dim.d_date_sk
      and web_sales.ws_bill_customer_sk = customer.c_customer_sk
      and d_month_seq between 1190 and 1190  + 11
  ) hot_cust
LIMIT
  100;
