-- SQLBench-DS query 9 derived from TPC-DS query 9 under the terms of the TPC Fair Use Policy.
-- TPC-DS queries are Copyright 2021 Transaction Processing Performance Council.
-- This query was generated at scale factor 1.
select
  case
    when (
      select
        count(*)
      from
        store_sales
      where
        ss_quantity between 1 and 20
    ) > 31002 then (
      select
        avg(ss_ext_discount_amt)
      from
        store_sales
      where
        ss_quantity between 1 and 20
    )
    else (
      select
        avg(ss_net_profit)
      from
        store_sales
      where
        ss_quantity between 1 and 20
    )
  end bucket1,
  case
    when (
      select
        count(*)
      from
        store_sales
      where
        ss_quantity between 21 and 40
    ) > 588 then (
      select
        avg(ss_ext_discount_amt)
      from
        store_sales
      where
        ss_quantity between 21 and 40
    )
    else (
      select
        avg(ss_net_profit)
      from
        store_sales
      where
        ss_quantity between 21 and 40
    )
  end bucket2,
  case
    when (
      select
        count(*)
      from
        store_sales
      where
        ss_quantity between 41 and 60
    ) > 2456 then (
      select
        avg(ss_ext_discount_amt)
      from
        store_sales
      where
        ss_quantity between 41 and 60
    )
    else (
      select
        avg(ss_net_profit)
      from
        store_sales
      where
        ss_quantity between 41 and 60
    )
  end bucket3,
  case
    when (
      select
        count(*)
      from
        store_sales
      where
        ss_quantity between 61 and 80
    ) > 21645 then (
      select
        avg(ss_ext_discount_amt)
      from
        store_sales
      where
        ss_quantity between 61 and 80
    )
    else (
      select
        avg(ss_net_profit)
      from
        store_sales
      where
        ss_quantity between 61 and 80
    )
  end bucket4,
  case
    when (
      select
        count(*)
      from
        store_sales
      where
        ss_quantity between 81 and 100
    ) > 20553 then (
      select
        avg(ss_ext_discount_amt)
      from
        store_sales
      where
        ss_quantity between 81 and 100
    )
    else (
      select
        avg(ss_net_profit)
      from
        store_sales
      where
        ss_quantity between 81 and 100
    )
  end bucket5
from
  reason
where
  r_reason_sk = 1;
