-- SQLBench-H query 15 derived from TPC-H query 15 under the terms of the TPC Fair Use Policy.
-- TPC-H queries are Copyright 1993-2022 Transaction Processing Performance Council.
create view revenue0 (supplier_no, total_revenue) as
select
  l_suppkey,
  sum(l_extendedprice * (1 - l_discount))
from
  lineitem
where
  l_shipdate >= date '1996-08-01'
  and l_shipdate < date '1996-08-01' + interval '3' month
group by
  l_suppkey;

select
  s_suppkey,
  s_name,
  s_address,
  s_phone,
  total_revenue
from
  supplier,
  revenue0
where
  s_suppkey = supplier_no
  and total_revenue = (
    select
      max(total_revenue)
    from
      revenue0
  )
order by
  s_suppkey;

drop view revenue0;
