-- SQLBench-H query 14 derived from TPC-H query 14 under the terms of the TPC Fair Use Policy.
-- TPC-H queries are Copyright 1993-2022 Transaction Processing Performance Council.
select
  100.00 * sum(
    case
      when p_type like 'PROMO%' then l_extendedprice * (1 - l_discount)
      else 0
    end
  ) / sum(l_extendedprice * (1 - l_discount)) as promo_revenue
from
  lineitem,
  part
where
  l_partkey = p_partkey
  and l_shipdate >= date '1995-02-01'
  and l_shipdate < date '1995-02-01' + interval '1' month;
