-- SQLBench-DS query 43 derived from TPC-DS query 43 under the terms of the TPC Fair Use Policy.
-- TPC-DS queries are Copyright 2021 Transaction Processing Performance Council.
-- This query was generated at scale factor 1.
select
  s_store_name,
  s_store_id,
  sum(
    case
      when (d_day_name = 'Sunday') then ss_sales_price
      else null
    end
  ) sun_sales,
  sum(
    case
      when (d_day_name = 'Monday') then ss_sales_price
      else null
    end
  ) mon_sales,
  sum(
    case
      when (d_day_name = 'Tuesday') then ss_sales_price
      else null
    end
  ) tue_sales,
  sum(
    case
      when (d_day_name = 'Wednesday') then ss_sales_price
      else null
    end
  ) wed_sales,
  sum(
    case
      when (d_day_name = 'Thursday') then ss_sales_price
      else null
    end
  ) thu_sales,
  sum(
    case
      when (d_day_name = 'Friday') then ss_sales_price
      else null
    end
  ) fri_sales,
  sum(
    case
      when (d_day_name = 'Saturday') then ss_sales_price
      else null
    end
  ) sat_sales
from
  date_dim,
  store_sales,
  store
where
  d_date_sk = ss_sold_date_sk
  and s_store_sk = ss_store_sk
  and s_gmt_offset = -5
  and d_year = 2000
group by
  s_store_name,
  s_store_id
order by
  s_store_name,
  s_store_id,
  sun_sales,
  mon_sales,
  tue_sales,
  wed_sales,
  thu_sales,
  fri_sales,
  sat_sales
LIMIT
  100;
