-- SQLBench-DS query 84 derived from TPC-DS query 84 under the terms of the TPC Fair Use Policy.
-- TPC-DS queries are Copyright 2021 Transaction Processing Performance Council.
-- This query was generated at scale factor 1.
select
  c_customer_id as customer_id,
  coalesce(c_last_name, '') || ', ' || coalesce(c_first_name, '') as customername
from
  customer,
  customer_address,
  customer_demographics,
  household_demographics,
  income_band,
  store_returns
where
  ca_city = 'White Oak'
  and c_current_addr_sk = ca_address_sk
  and ib_lower_bound >= 45626
  and ib_upper_bound <= 45626 + 50000
  and ib_income_band_sk = hd_income_band_sk
  and cd_demo_sk = c_current_cdemo_sk
  and hd_demo_sk = c_current_hdemo_sk
  and sr_cdemo_sk = cd_demo_sk
order by
  c_customer_id
LIMIT
  100;
